@@ -5,13 +5,15 @@ import socket
 import subprocess
 from simpervisor import SupervisedProcess
 from datetime import datetime
-from asyncio import Lock
+from asyncio import Lock, ensure_future
 from .util import url_path_join
 from .websocket import WebSocketHandlerMixin, pingable_ws_connect
-from tornado.log import app_log
+from tornado.log import app_log, gen_log
 from jupyterhub.services.auth import HubOAuthenticated
-from urllib.parse import urlunparse, urlparse
-
+from urllib.parse import urlunparse, urlparse, quote
+from typing import (
+    Any
+)
 
 class AddSlashHandler(web.RequestHandler):
     """Add trailing slash to URLs that need them."""
@@ -36,6 +38,7 @@ class ProxyHandler(HubOAuthenticated, WebSocketHandlerMixin):
         self.proxy_base = ''
         self.absolute_url = kwargs.pop('absolute_url', False)
         self.host_whitelist = kwargs.pop('host_whitelist', ['localhost', '127.0.0.1'])
+        self.subprotocols = None
         super().__init__(*args, **kwargs)
 
     @property
@@ -133,6 +136,8 @@ class ProxyHandler(HubOAuthenticated, WebSocketHandlerMixin):
         else:
             client_path = proxied_path
 
+        client_path = quote(client_path, safe=":/?#[]@!$&'()*+,;=-._~")
+
         client_uri = '{protocol}://{host}:{port}{path}'.format(
             protocol=protocol,
             host=host,
@@ -204,7 +209,17 @@ class ProxyHandler(HubOAuthenticated, WebSocketHandlerMixin):
 
         req = self._build_proxy_request(host, port, proxied_path, body)
 
-        response = await client.fetch(req, raise_error=False)
+        try:
+            response = await client.fetch(req, raise_error=False)
+        except httpclient.HTTPError as err:
+            if err.code == 599:
+                self._record_activity()
+                self.set_status(599)
+                self.write(str(err))
+                return
+            else:
+                raise
+
         # record activity at start and end of requests
         self._record_activity()
 
@@ -227,11 +242,88 @@ class ProxyHandler(HubOAuthenticated, WebSocketHandlerMixin):
             if response.body:
                 self.write(response.body)
 
+    async def ws_get(self, *args: Any, **kwargs: Any) -> None:
+        """
+        A version of WebSocketHandler.get that also opens the underlying process' websocket so that
+        any GET headers can also be passed back to the client.
+        """
+        self.open_args = args
+        self.open_kwargs = kwargs
+
+        # Upgrade header should be present and should be equal to WebSocket
+        if self.request.headers.get("Upgrade", "").lower() != "websocket":
+            self.set_status(400)
+            log_msg = 'Can "Upgrade" only to "WebSocket".'
+            self.finish(log_msg)
+            gen_log.debug(log_msg)
+            return
+
+        # Connection header should be upgrade.
+        # Some proxy servers/load balancers
+        # might mess with it.
+        headers = self.request.headers
+        connection = map(
+            lambda s: s.strip().lower(), headers.get("Connection", "").split(",")
+        )
+        if "upgrade" not in connection:
+            self.set_status(400)
+            log_msg = '"Connection" must be "Upgrade".'
+            self.finish(log_msg)
+            gen_log.debug(log_msg)
+            return
+
+        # Handle WebSocket Origin naming convention differences
+        # The difference between version 8 and 13 is that in 8 the
+        # client sends a "Sec-Websocket-Origin" header and in 13 it's
+        # simply "Origin".
+        if "Origin" in self.request.headers:
+            origin = self.request.headers.get("Origin")
+        else:
+            origin = self.request.headers.get("Sec-Websocket-Origin", None)
+
+        # If there was an origin header, check to make sure it matches
+        # according to check_origin. When the origin is None, we assume it
+        # did not come from a browser and that it can be passed on.
+        if origin is not None and not self.check_origin(origin):
+            self.set_status(403)
+            log_msg = "Cross origin websockets not allowed"
+            self.finish(log_msg)
+            gen_log.debug(log_msg)
+            return
+
+        # Now open connection to underlying web process from ourself - BESPOKE
+        path = ''
+        if len(args) > 0:
+            path = args[0]
+        elif 'path' in kwargs:
+            path = kwargs['path']
+
+        r = await self._ws_open_proxy('localhost', self.port, path)
+
+        if hasattr(self, 'handshake_headers'):
+            # Any headers need passing on?
+            for header, v in self.handshake_headers.get_all():
+                if header in ('Set-Cookie', 'Vary'):
+                    # some header appear multiple times, eg 'Set-Cookie'
+                    self.add_header(header, v)
+
+        # Now establish websocket between client and ourself - BACK TO ORIGINAL CODE NOW
+        self.ws_connection = self.get_websocket_protocol()
+        if self.ws_connection:
+            await self.ws_connection.accept_connection(self)
+        else:
+            self.set_status(426, "Upgrade Required")
+            self.set_header("Sec-WebSocket-Version", "7, 8, 13")
+
     async def proxy_open(self, host, port, proxied_path=''):
         """
         Called when a client opens a websocket connection.
-        We establish a websocket connection to the proxied backend &
-        set up a callback to relay messages through.
+        """
+        pass
+
+    async def _ws_open_proxy(self, host, port, proxied_path=''):
+        """
+        Open the websocket of the underlying process.
         """
 
         if not self._check_host_whitelist(host):
@@ -249,6 +341,9 @@ class ProxyHandler(HubOAuthenticated, WebSocketHandlerMixin):
         current_loop = ioloop.IOLoop.current()
         ws_connected = current_loop.asyncio_loop.create_future()
 
+        def headers_cb(headers):
+            self.handshake_headers = headers
+
         def message_cb(message):
             """
             Callback when the backend sends messages to us
@@ -275,7 +370,9 @@ class ProxyHandler(HubOAuthenticated, WebSocketHandlerMixin):
             self._record_activity()
             request = httpclient.HTTPRequest(url=client_uri, headers=headers)
             self.ws = await pingable_ws_connect(request=request,
-                                                on_message_callback=message_cb, on_ping_callback=ping_cb)
+                                                on_message_callback=message_cb, on_ping_callback=ping_cb,
+                                                on_get_headers_callback=headers_cb,
+                                                subprotocols=self.subprotocols)
             ws_connected.set_result(True)
             self._record_activity()
             self.log.info('Websocket connection established to {}'.format(client_uri))
@@ -306,6 +403,7 @@ class ProxyHandler(HubOAuthenticated, WebSocketHandlerMixin):
 
     def select_subprotocol(self, subprotocols):
         '''Select a single Sec-WebSocket-Protocol during handshake.'''
+        self.subprotocols = subprotocols
         if isinstance(subprotocols, list) and subprotocols:
             self.log.info('Client sent subprotocols: {}'.format(subprotocols))
             return subprotocols[0]
@@ -407,6 +505,21 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
 {stdout}
     </pre>
     
+    </body>
+</html>"""
+
+    gitpulling_template = """<!DOCTYPE html>
+<html>
+    <head>
+        <title>Status page from ContainDS Dashboards</title>
+        <meta http-equiv="Refresh" content="2">
+    </head>
+    <body>
+
+    <h3>Pulling content from git</h2>
+
+    <p>Please wait.... refreshing page in 2 seconds</p>
+    
     </body>
 </html>"""
 
@@ -417,6 +530,10 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
         self.stderr_str = None
         self.stdout_str = None
 
+        self.origin_host = None
+
+        self.ready_check_path = '/'
+
         super().__init__(*args, **kwargs)
 
     def initialize(self, state, authtype, *args, **kwargs):
@@ -462,7 +579,7 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
         return 5
 
     async def _http_ready_func(self, p):
-        url = 'http://localhost:{}'.format(self.port)
+        url = 'http://localhost:{}{}'.format(self.port, self.ready_check_path)
         async with aiohttp.ClientSession() as session:
             try:
                 async with session.get(url) as resp:
@@ -507,9 +624,6 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
 
                     if not is_ready:
 
-                        #if proc._restart_process_future:
-                        #    proc._restart_process_future.cancel()
-
                         self.stderr_str = None
                         self.stdout_str = None
 
@@ -527,6 +641,24 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
 
                         del self.state['proc']
                         return False
+
+                    else:
+                        # Make sure we empty the buffers periodically
+
+                        async def pipe_output(proc, pipename, log):
+                            while True:
+                                if proc.proc:
+                                    stream = getattr(proc.proc, pipename, None)
+                                    if stream:
+                                        line = await stream.readline()
+                                        if line:
+                                            log.info(line)
+                                        else:
+                                            break
+
+                        ensure_future(pipe_output(proc, 'stderr', self.log))
+                        ensure_future(pipe_output(proc, 'stdout', self.log))
+
                 except:
                     # Make sure we remove proc from state in any error condition
                     del self.state['proc']
@@ -538,6 +670,11 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
         return await self.core_proxy(port, path)
 
     async def core_proxy(self, port, path):
+
+        if self.origin_host is None:
+            # Get origin from this request
+            self.store_origin_host()
+
         if not path.startswith('/'):
             path = '/' + path
 
@@ -547,6 +684,20 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
             else:
                 path = self.mappath.get(path, path)
 
+        if self.gitwrapper:
+            if not self.gitwrapper.finished:
+                self.set_status(200)
+                return self.write(self.gitpulling_template)
+            elif self.gitwrapper.error:
+                from tornado.escape import xhtml_escape
+                html = self.error_template.format(
+                    cmd=xhtml_escape(" ".join(self.get_cmd())),
+                    stderr=xhtml_escape("\n".join(self.gitwrapper.logs)),
+                    stdout=''
+                )
+                self.set_status(500)
+                return self.write(html)
+
         if not await self.ensure_process():
             from tornado.escape import xhtml_escape
             html = self.error_template.format(
@@ -570,6 +721,13 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
         return await self.proxy(self.port, path)
 
     async def open(self, path):
+        if self.origin_host is None:
+            # Get origin from this request
+            self.store_origin_host()
+        
+        if self.gitwrapper and not self.gitwrapper.finished:
+            raise web.HTTPError(500, 'Git checkout is not finished')
+
         if await self.ensure_process():
             return await super().open(self.port, path)
         raise web.HTTPError(500, 'could not start {} in time'.format(self.name))
@@ -592,8 +750,12 @@ class SuperviseAndProxyHandler(LocalProxyHandler):
     def options(self, path):
         return self.proxy(self.port, path)
 
+    def store_origin_host(self):
+        self.log.debug('Storing origin host {}'.format(self.request.host))
+        self.origin_host = self.request.host
 
-def _make_serverproxy_handler(name, command, environment, timeout, absolute_url, port, mappath):
+
+def _make_serverproxy_handler(name, command, environment, timeout, absolute_url, port, ready_check_path, gitwrapper, mappath):
     """
     Create a SuperviseAndProxyHandler subclass with given parameters
     """
@@ -606,19 +768,38 @@ def _make_serverproxy_handler(name, command, environment, timeout, absolute_url,
             self.absolute_url = absolute_url
             self.requested_port = port
             self.mappath = mappath
+            self.ready_check_path = ready_check_path
+            self.gitwrapper = gitwrapper
 
         @property
         def process_args(self):
             return {
                 'port': self.port,
                 'base_url': self.base_url,
+                'presentation_path': self.presentation_path,
+                'presentation_basename': self.presentation_basename,
+                'presentation_dirname': self.presentation_dirname,
+                'origin_host': self.origin_host,
                 '-': '-',
                 '--': '--'
             }
+
         @property
         def base_url(self):
             return self.settings.get('base_url', '/')
 
+        @property
+        def presentation_path(self):
+            return self.settings.get('presentation_path', '.')
+
+        @property
+        def presentation_basename(self):
+            return self.settings.get('presentation_basename', '')
+
+        @property
+        def presentation_dirname(self):
+            return self.settings.get('presentation_dirname', '.')
+
         @property
         def hub_users(self):
             return {self.settings['user']}
